@@ -9,7 +9,6 @@ use quote::quote;
 use quote::quote_spanned;
 use quote::ToTokens;
 use syn::parse_macro_input;
-use syn::punctuated::Punctuated;
 use syn::spanned::Spanned;
 use syn::token::Brace;
 use syn::token::Gt;
@@ -35,7 +34,7 @@ use syn::TypeParam;
 use syn::TypeParamBound;
 use syn::TypePath;
 use syn::Visibility;
-use syn_utils::TypeMatcher;
+use syn_utils::TypeOrPath;
 
 use crate::parse_assoc_type::parse_assoc_type;
 use crate::parse_assoc_type::AssocTypeParseError;
@@ -43,7 +42,7 @@ use crate::parse_trait_sig::parse_trait_signature;
 use crate::parse_trait_sig::MethodParseError;
 use crate::parse_trait_sig::SignatureChanges;
 use crate::parse_trait_sig::TypeTransform;
-use crate::syn_utils::find_in_path;
+use crate::syn_utils::iter_path;
 use crate::syn_utils::trait_bounds;
 use crate::transform::AssocTypeConversions;
 
@@ -97,7 +96,7 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
                 Err((span, MethodParseError::UnconvertibleAssocType)) => {
                     return abort!(
                         span,
-                        "associated type is either undefined or doesn't have an Into bound"
+                        "associated type is either undefined or doesn't have a trait bound"
                     )
                 }
                 Err((span, MethodParseError::AssocTypeInInputs)) => {
@@ -111,9 +110,11 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
                     MethodParseError::AssocTypeInUnsupportedReturnType
                     | MethodParseError::UnconvertibleAssocTypeInFnInput,
                 )) => return abort!(span, "dynamize does not know how to convert this type"),
-                Err((span, MethodParseError::UnconvertibleAssocTypeInTraitBound)) => {
-                    return abort!(span, "dynamize does not support associated types here")
-                }
+                Err((
+                    span,
+                    MethodParseError::UnconvertibleAssocTypeInTraitBound
+                    | MethodParseError::UnconvertibleAssocTypeInWhereClause,
+                )) => return abort!(span, "dynamize does not support associated types here"),
                 Err((span, MethodParseError::ImplTraitInInputs)) => {
                     return abort!(
                         span,
@@ -171,23 +172,30 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
         trait_token: Trait::default(),
         generics: original_trait.generics.clone(),
         colon_token: None,
-        supertraits: Punctuated::new(),
+        supertraits: original_trait
+            .supertraits
+            .iter()
+            .filter(|t| match t {
+                TypeParamBound::Trait(t) => !t.path.is_ident("Sized"),
+                TypeParamBound::Lifetime(_) => true,
+            })
+            .cloned()
+            .collect(),
         brace_token: Brace::default(),
         items: Vec::new(),
     };
 
     let mut generic_map = HashMap::new();
 
-    for generic in &dyn_trait.generics.params {
-        if let GenericParam::Type(type_param) = generic {
-            generic_map.insert(type_param.ident.clone(), type_param.bounds.clone());
-            for trait_bound in trait_bounds(&type_param.bounds) {
-                if let Some(assoc_type) = find_in_path(&trait_bound.path, &AssocTypeMatcher) {
-                    return abort!(
-                        assoc_type.span(),
-                        "dynamize does not support associated types in trait generic bounds"
-                    );
-                }
+    for type_param in dyn_trait.generics.type_params() {
+        generic_map.insert(type_param.ident.clone(), type_param.bounds.clone());
+        for trait_bound in trait_bounds(&type_param.bounds) {
+            if let Some(assoc_type) = iter_path(&trait_bound.path).find_map(filter_map_assoc_paths)
+            {
+                return abort!(
+                    assoc_type.span(),
+                    "dynamize does not support associated types in trait generic bounds"
+                );
             }
         }
     }
@@ -302,12 +310,18 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
 
     let blanket_impl = generate_blanket_impl(&dyn_trait, &original_trait, &method_impls);
 
+    let dyn_trait_name = &dyn_trait.ident;
+    let (impl_generics, ty_generics, where_clause) = dyn_trait.generics.split_for_impl();
+
     let expanded = quote! {
         #original_trait
 
         #(#dyn_trait_attrs)*
         #dyn_trait
 
+        // assert that dyn_trait can actually be made into an object
+        impl #impl_generics dyn #dyn_trait_name #ty_generics #where_clause {}
+
         #(#blanket_impl_attrs)*
         #blanket_impl
     };
@@ -338,36 +352,30 @@ fn generate_blanket_impl(
         eq_token: None,
         default: None,
     }));
-    let (_, type_gen, _where) = dyn_trait.generics.split_for_impl();
+    let (_, type_gen, where_clause) = dyn_trait.generics.split_for_impl();
     let dyn_trait_name = &dyn_trait.ident;
     quote! {
-        impl #blanket_generics #dyn_trait_name #type_gen for #some_ident {
+        impl #blanket_generics #dyn_trait_name #type_gen for #some_ident #where_clause {
             #(#method_impls)*
         }
     }
 }
 
-struct AssocTypeMatcher;
-impl TypeMatcher<Path> for AssocTypeMatcher {
-    fn match_path<'a>(&self, path: &'a Path) -> Option<&'a Path> {
-        if path.segments.first().unwrap().ident == "Self" {
-            return Some(path);
-        }
-        None
-    }
+fn path_is_assoc_type(path: &Path) -> bool {
+    path.segments.first().unwrap().ident == "Self"
 }
 
-impl<TypeMatch, PathMatch, T> TypeMatcher<T> for (TypeMatch, PathMatch)
-where
-    TypeMatch: Fn(&Type) -> Option<&T>,
-    PathMatch: Fn(&Path) -> Option<&T>,
-{
-    fn match_type<'a>(&self, t: &'a Type) -> Option<&'a T> {
-        self.0(t)
+fn match_assoc_type(item: TypeOrPath) -> bool {
+    if let TypeOrPath::Path(path) = item {
+        return path_is_assoc_type(path);
     }
+    false
+}
 
-    fn match_path<'a>(&self, path: &'a Path) -> Option<&'a T> {
-        self.1(path)
+fn filter_map_assoc_paths(item: TypeOrPath) -> Option<&Path> {
+    match item {
+        TypeOrPath::Path(p) if path_is_assoc_type(p) => Some(p),
+        _other => None,
     }
 }
 
@@ -382,6 +390,10 @@ impl TypeTransform {
                 let inner = opt.convert(quote!(x));
                 quote! {#arg.map(|x| #inner)}
             }
+            TypeTransform::Iterator(box_type, inner) => {
+                let inner = inner.convert(quote!(x));
+                quote! {Box::new(#arg.map(|x| #inner)) as #box_type}
+            }
             TypeTransform::Result(ok, err) => {
                 let map_ok = !matches!(ok.as_ref(), TypeTransform::NoOp);
                 let map_err = !matches!(err.as_ref(), TypeTransform::NoOp);
@@ -402,42 +414,6 @@ impl TypeTransform {
     }
 }
 
-/// Just a convenience trait for us to avoid match/if-let blocks everywhere.
-trait As<T> {
-    fn get_as(&self) -> Option<&T>;
-    fn get_as_mut(&mut self) -> Option<&mut T>;
-}
-
-impl As<AngleBracketedGenericArguments> for PathArguments {
-    fn get_as(&self) -> Option<&AngleBracketedGenericArguments> {
-        match self {
-            PathArguments::AngleBracketed(args) => Some(args),
-            _other => None,
-        }
-    }
-    fn get_as_mut(&mut self) -> Option<&mut AngleBracketedGenericArguments> {
-        match self {
-            PathArguments::AngleBracketed(args) => Some(args),
-            _other => None,
-        }
-    }
-}
-
-impl As<Type> for GenericArgument {
-    fn get_as(&self) -> Option<&Type> {
-        match self {
-            GenericArgument::Type(typearg) => Some(typearg),
-            _other => None,
-        }
-    }
-    fn get_as_mut(&mut self) -> Option<&mut Type> {
-        match self {
-            GenericArgument::Type(typearg) => Some(typearg),
-            _other => None,
-        }
-    }
-}
-
 fn path_segment_for_trait(sometrait: &ItemTrait) -> PathSegment {
     PathSegment {
         ident: sometrait.ident.clone(),
@@ -468,3 +444,42 @@ fn path_segment_for_trait(sometrait: &ItemTrait) -> PathSegment {
         },
     }
 }
+
+#[doc = include_str!("../tests/doctests.md")]
+#[cfg(doctest)]
+struct Doctests;
+
+#[test]
+fn ui() {
+    use std::process::Command;
+
+    for entry in std::fs::read_dir("ui-tests/src/bin").unwrap().flatten() {
+        if entry.path().extension().unwrap() == "rs" {
+            let output = Command::new("cargo")
+                .arg("check")
+                .arg("--quiet")
+                .arg("--bin")
+                .arg(entry.path().file_stem().unwrap())
+                .current_dir("ui-tests")
+                .output()
+                .unwrap();
+            let received_stderr = std::str::from_utf8(&output.stderr)
+                .unwrap()
+                .lines()
+                .filter(|l| !l.starts_with("error: could not compile"))
+                .collect::<Vec<_>>()
+                .join("\n");
+            let expected_stderr =
+                std::fs::read_to_string(entry.path().with_extension("stderr")).unwrap_or_default();
+            if received_stderr != expected_stderr {
+                println!(
+                    "EXPECTED:\n{banner}\n{expected}{banner}\n\nACTUAL OUTPUT:\n{banner}\n{actual}{banner}",
+                    banner = "-".repeat(30),
+                    expected = expected_stderr,
+                    actual = received_stderr
+                );
+                panic!("failed");
+            }
+        }
+    }
+}
