@@ -6,10 +6,10 @@ use syn::{
 };
 use syn::{Ident, Signature, TypeImplTrait};
 
+use crate::match_assoc_type;
 use crate::parse_assoc_type::BoxType;
-use crate::syn_utils::{find_in_type, trait_bounds, TypeMatcher};
+use crate::syn_utils::{iter_type, trait_bounds, TypeOrPath};
 use crate::transform::{dynamize_function_bounds, AssocTypeConversions, TransformError};
-use crate::AssocTypeMatcher;
 
 #[derive(Debug, Clone)]
 pub enum TypeTransform {
@@ -17,6 +17,7 @@ pub enum TypeTransform {
     Into,
     Box(BoxType),
     Map(Box<TypeTransform>),
+    Iterator(BoxType, Box<TypeTransform>),
     Result(Box<TypeTransform>, Box<TypeTransform>),
 }
 
@@ -28,17 +29,14 @@ pub enum MethodParseError {
     AssocTypeInUnsupportedReturnType,
     UnconvertibleAssocTypeInFnInput,
     UnconvertibleAssocTypeInTraitBound,
+    UnconvertibleAssocTypeInWhereClause,
     UnconvertibleAssocType,
 }
 
-struct ImplTraitMatcher;
-impl TypeMatcher<TypeImplTrait> for ImplTraitMatcher {
-    fn match_type<'a>(&self, t: &'a Type) -> Option<&'a TypeImplTrait> {
-        if let Type::ImplTrait(impltrait) = t {
-            Some(impltrait)
-        } else {
-            None
-        }
+fn filter_map_impl_trait(item: TypeOrPath) -> Option<&TypeImplTrait> {
+    match item {
+        TypeOrPath::Type(Type::ImplTrait(impltrait)) => Some(impltrait),
+        _other => None,
     }
 }
 
@@ -58,10 +56,10 @@ pub fn parse_trait_signature(
     // provide better error messages for associated types in params
     for input in &signature.inputs {
         if let FnArg::Typed(pattype) = input {
-            if find_in_type(&pattype.ty, &AssocTypeMatcher).is_some() {
+            if iter_type(&pattype.ty).any(match_assoc_type) {
                 return Err((pattype.ty.span(), MethodParseError::AssocTypeInInputs));
             }
-            if let Some(impl_trait) = find_in_type(&pattype.ty, &ImplTraitMatcher) {
+            if let Some(impl_trait) = iter_type(&pattype.ty).find_map(filter_map_impl_trait) {
                 return Err((impl_trait.span(), MethodParseError::ImplTraitInInputs));
             }
         }
